#==============================================================================
# mwutil, scaleutil, and scaleutilmisc package index file.
#
# Copyright (c) 2020-2024  Csaba Nemethi (E-mail: csaba.nemethi@t-online.de)
#==============================================================================

package ifneeded mwutil     2.23   [list source [file join $dir mwutil.tcl]]
package ifneeded scaleutil  1.14.1 [list source [file join $dir scaleutil.tcl]]
package ifneeded themepatch 1.8    [list source [file join $dir themepatch.tcl]]
package ifneeded scaleutilmisc 1.7.1 \
	[list source [file join $dir scaleutilMisc.tcl]]
